/* file_ro.c */
/*
 * For !OWW project
 * One-wire weather
 * Dr. Simon J. Melhuish
 * 1999 - 2000
 * Free for non-comercial use
 * Dallas parts subject to their copyright and conditions
 *
 */
/* Linux specific file functions for Oww
   Simon J. Melhuish
   Tue 26th September 2000
*/
#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

//#ifdef _WIN32
//#  include "win32config.h"
//#endif


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
//#include <glib.h>
#include <fcntl.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif

#include <errno.h>
#include "file.h"
//extern char *setup_format_loghead ;
void owwlog_warning(char* format, ...) ;
void owwlog_error(char* format, ...) ;

static int file_create_dir(char *dirname)
{
#ifdef WIN32
  return (mkdir(dirname)) ;
#else
  return (mkdir(dirname, 0755)) ;
#endif
}
static int file_create_file(char *filename, int type)
{
  /* Try to create the file */
  int fd ;
  /* We don't use the RISC OS filetype */
  type = type ;
  /*werr(0,"file_create_file(%s, %d)", filename, type) ;*/
  fd = creat(filename, 0664) ;
  if (fd == -1) {
    /* Open failed */
    return FILE_CREATE_ERROR ;
  }
  close(fd) ;
  return FILE_CREATE_OK ;
}
int file_check_dir(char *dir_name, int create)
{
  /* Check that directory dir_name exists */
  struct stat statbuf ;
  /*werr(0, "file_check_dir(%s, %d)", dir_name, create) ;*/
  /* Check that directory exists */
  if (access(dir_name, F_OK)) {
    /* Directory does not exist */
    if (!create)
    {
      owwlog_warning("%s not found.", dir_name) ;
      return FILE_CHECK_ERROR ;
    }
    else
    {
      if (file_create_dir(dir_name) != 0) 
      {
        owwlog_warning("Failed to create %s", dir_name) ;
        return FILE_CHECK_ERROR ;
      }
      return FILE_CHECK_OK ;
    }
  }
  /* dir_name exists, but is it a directory? */
  if (stat(dir_name, &statbuf)) {
    /* stat failed */
    owwlog_warning("Unable to get directory information: %s",
      strerror(errno)) ;
    return FILE_CHECK_ERROR ;
  }
  /* statbuf.st_mode tells us the mode */
  if (!S_ISDIR(statbuf.st_mode)) {
    /* dir_name is something other than a directory */
    owwlog_warning("%s is not a directory", dir_name) ;
    return FILE_CHECK_ERROR ;
  }
  /* Do we have write permission? */
  if (access(dir_name, W_OK)) {
    /* Directory does not give us write permission */
    owwlog_warning("No write permission in %s", dir_name) ;
    return FILE_CHECK_ERROR ;
  }
  return FILE_CHECK_OK ; /* OK */
}

int file_check_file(char *file_name, int create, FileMaker func, void *user_data)
{
  /* Check that file file_name exists */
  struct stat statbuf ;
  /*werr(0, "file_check_file(%s, %d)", file_name, create) ;*/
  /* Check that file exists */
  if (access(file_name, F_OK)) {
    /* File does not exist */
    if (!create)
    {
      owwlog_warning("File %s not found", file_name) ;
      return FILE_CHECK_ERROR ;
    }
    else
    {
      //FILE *fp ;
      if (file_create_file(file_name, 0xdfe) != FILE_CREATE_OK) {
        owwlog_warning("Failed to create file %s", file_name) ;
        return FILE_CHECK_ERROR ;
      }
      
      if (func) return func(file_name, user_data) ;
      //fp = fopen(file_name, "w") ;
      //if (!fp) return -1 ;
      //fprintf(fp, setup_format_loghead) ;
      //fprintf(fp, "\n") ;
      //fclose(fp) ;
      return FILE_CHECK_OK ;
    }
  }

  /* file_name exists, but is it a file? */
  if (stat(file_name, &statbuf)) {
    /* stat failed */
    owwlog_warning("Unable to get file information: %s",
      strerror(errno)) ;
    return FILE_CHECK_ERROR ;
  }
  /* statbuf.st_mode tells us the mode */
  if (!S_ISREG(statbuf.st_mode)) {
    /* file_name is something other than a directory */
    owwlog_warning("%s is not a normal file", file_name) ;
    return FILE_CHECK_ERROR ;
  }
  /* Do we have write permission? */
  if (access(file_name, W_OK)) {
    /* File does not give us write permission */
    owwlog_warning("No write permission in %s", file_name) ;
    return FILE_CHECK_ERROR ;
  }
  return FILE_CHECK_OK ; /* OK */
}
