/* logger.h
 *
 * Simon Melhuish
 * Cardiff Astronomy Instrumentation Group
 * 2004
 *
 * Logger class for owwlog
 */
 
#ifndef LOGGER_H
#define LOGGER_H 1
 
#include "owwl.h"

#define LOGGER_ERROR -1
#define LOGGER_OK 0
 
/* Function types */

typedef int (* LoggerFunc) (owwl_data *data, void *user_data) ;
typedef int (* LoggerReadoutFunc) (owwl_data *data, void *user_data, double *result, int arg) ;


typedef struct
{
  LoggerFunc reset       ; /* Reset the internal data structure */
  LoggerFunc accumulate  ; /* Accumulate new values to internal data structure */
  LoggerFunc calculate   ; /* Readout to internal data structure */
  LoggerReadoutFunc result ; /* Readout a result as a double */
  
  int results ; /* No. of results available */
  int N ;
  void *internal ;
  void *calib ;
} Logger ;

typedef struct
{
  int count ;
  double *sum ;
} LoggerFloats ;

typedef struct
{
  double x_sum ;
  double y_sum ;
  double s_sum ;
  double s_max ;
} LoggerWind ;

typedef struct
{
  unsigned long c ;
  double r ;
  double r_sum ;
  time_t t ;
} LoggerRain ;

typedef struct
{
  unsigned long c ;
  long d ;
  double r ;
  double r_sum ;
  time_t t ;
} LoggerGpc ;

int logger_reset         (owwl_conn *conn, owwl_data *data, void *user_data) ;

int logger_accumulate    (owwl_conn *conn, owwl_data *data, void *user_data) ;

int logger_calculate     (owwl_conn *conn, owwl_data *data, void *user_data) ;

int logger_result        (owwl_data *data, void *user_data, double *result, int arg) ;

int logger_result_string (owwl_data *data, void *user_data, int arg, char *buffer, int bufflen, char *format) ;

void logger_build_data   (owwl_conn *conn, owwl_data *data, void *user_data) ;

int logger_result_name   (owwl_data *data, void *user_data, int arg, char *buffer, int bufflen) ;

#endif
