
/* owwlog
 *
 * Simon Melhuish
 * Cardiff Astronomy Instrumentation Group
 *
 * 2004
 */

#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

//#ifdef _WIN32
//#  include "win32/config.h"
//#endif

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <popt.h>
/* POPT_TABLEEND is only defined from popt 1.6.1 */
#ifndef POPT_TABLEEND
# define POPT_TABLEEND { NULL, '\0', 0, 0, 0, NULL, NULL }
#endif

#include <stdarg.h>
#include <sockdef.h>
#ifdef HAVE_ARPA_INET_H
#include <arpa/inet.h>
#endif

//#include <glib.h>

#include <sys/types.h>

#ifdef HAVE_SYS_WAIT_H
#include <sys/wait.h>
#endif

#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#endif

#include <sys/file.h>

#include "owwl.h"
#include "logger.h"
#include "file.h"
#include "calib.h"

#define MAXFILENAME 128
#define MAXLINE 1024

#ifdef _WIN32
char setup_format_logfile[] = "\\%y%m\\%y%m%d.txt" ;
char setup_format_logdir[] = "\\%y%m" ;
#else
char setup_format_logfile[] = "/%y%m/%y%m%d.txt" ;
char setup_format_logdir[] = "/%y%m" ;
//char setup_format_loghead[] =  ;
#endif

int opt_daemon = 0 ;
int opt_port = 8899 ;
int opt_broadcast = 8899 ;
int opt_broadcaster = 0 ;
char *opt_server = NULL ;
char *opt_execute = NULL ;
char *opt_logfile = NULL ;
char *opt_logdir = NULL ;
char *opt_livefile = NULL ;
char *opt_ignore = NULL ;
char *opt_rrdfile = NULL ;
char *opt_caldir = NULL ;
int opt_time = 60 ;
int opt_abort = 30 ;
int opt_quiet = 0 ;
int opt_iso_time = 0 ;
int opt_unix_time = 0 ;
int opt_log_snap = 0 ;

int data_updated ;

char **ignored_vec = NULL ;
int ignore_count = 0 ;
int run_quiet = 0 ;

static time_t the_time = 0, log_time = 0 ;

char rrd_template[MAXLINE] ;

/* Protoype for test function */

int
owwl_test(owwl_conn *conn) ;

/* Messages, warnings and errors */
void owwlog_message(char* format, ...) ;
void owwlog_warning(char* format, ...) ;
void owwlog_error(char* format, ...) ;

void reduce_name(char *name) ;

void owwlog_message(char* format, ...)
{
  va_list va;
  
  if (run_quiet) return ;

  va_start(va, format);
  vfprintf(stdout, format, va);
  va_end(va);
}

void owwlog_warning(char* format, ...)
{
  va_list va;
  char message[100] ;

  strftime(message, sizeof(message), "[%y-%m-%d %H:%M:%S] Warning: ", localtime(&the_time)) ;
  fprintf(stderr, message) ;
  va_start(va, format);
  vfprintf(stderr, format, va);
  va_end(va);
}

void owwlog_error(char* format, ...)
{
  va_list va;

  char message[100] ;

  strftime(message, sizeof(message), "[%y-%m-%d %H:%M:%S] Error: ", localtime(&the_time)) ;
  fprintf(stderr, message) ;
  va_start(va, format);
  vfprintf(stderr, format, va);
  va_end(va);
#ifdef _WIN32
  WSACleanup();
#endif
  exit(0) ;
}

static SOCKET
client_connect(char *hostname, int port)
{
  /* Try to connect to the client */
  
  SOCKET socket_client ;
  struct hostent *host ;
  u_long a = 1L ;

  
  // Address structures are static, so that we can reuse them for reconnecting
  static int addr_len = 0 ;
  static struct sockaddr *address = NULL ;
# ifdef HAVE_SYS_UN_H
  static struct sockaddr_un addr_un ;
# endif
# ifdef HAVE_NETINET_IN_H
  static struct sockaddr_in addr_in ;
# endif

  if (address == NULL)
  {
    /* What kind of connexion? */
  
    if (hostname[0] == '/') /* AF_LOCAL */
    {
# ifdef HAVE_SYS_UN_H
      memset(&addr_un, 0, sizeof(addr_un)) ;
  
      addr_un.sun_family = AF_LOCAL ;
      strcpy(addr_un.sun_path, hostname) ;
      addr_len = sizeof(addr_un.sun_family) + strlen(hostname) ;
      address = (struct sockaddr *) &addr_un ;
      owwlog_message("Connecting to LOCAL (UNIX) port: %s\n", hostname) ;
# else
      return INVALID_SOCKET ;
# endif // HAVE_SYS_UN_H
    }
    else
    {
# ifdef HAVE_NETINET_IN_H
      /* Just use the ordinary (blocking) dns lookup for now */
      owwlog_message("Looking up \"%s\"\n", hostname) ;
  
      host = gethostbyname(hostname) ;
  
      if (!host)
      {
        owwlog_warning("Unable to resolve host name \"%s\"\n", hostname) ;
        return INVALID_SOCKET ;
      }
  
      memset(&addr_in, 0, sizeof(struct sockaddr_in));
      
      //addr_in.sin_family = AF_INET ;
      //addr_in.sin_port   = htons(port) ;
  
      /* Take the first ip address */
      
      memcpy(&addr_in.sin_addr.s_addr, host->h_addr, host->h_length);
      addr_in.sin_family = host->h_addrtype;
      addr_in.sin_port   = htons(port) ;
      owwlog_message("Connecting by IP to server %s:%d\n", hostname, port) ;
      address = (struct sockaddr *) &addr_in ;
      addr_len = sizeof(addr_in) ;
  
  /*    switch (host->h_addrtype)
      {
        case AF_INET:
        {
          memcpy(&addr_in.sin_addr,
                host->h_addr_list[0],
                sizeof(addr_in.sin_addr)) ;
          address = (struct sockaddr *) &addr_in ;
          addr_len = sizeof(addr_in) ;
          owwlog_message("Connecting by IP to server %s:%d\n", hostname, port) ;
  #ifdef HAVE_ARPA_INET_H
          owwlog_message("Address: %s\n", inet_ntoa(addr_in.sin_addr)) ;
  #endif
          break ;
        }
        
        default:
        {
          owwlog_warning("Address type %d not supported\n", host->h_addrtype) ;
          return INVALID_SOCKET ;
        }
      }*/
# else
      return INVALID_SOCKET ;
# endif // ifdef HAVE_NETINET_IN_H
  
    }
  }

  socket_client = socket(address->sa_family, SOCK_STREAM, 0) ;

  if (socket_client == INVALID_SOCKET)
  {
    owwlog_warning("Unable to open socket for communication with \"%s\"\n",
                   hostname) ;
    return INVALID_SOCKET ;
  }

  /* Now connect to the server */

  if (connect(socket_client,
              address, addr_len))
  {
    owwlog_warning("Unable to connect to \"%s\"\n",
                   hostname) ;
    CLOSESOCK(socket_client);
    return INVALID_SOCKET ;
  }

  /* Connected Ok */

  /* Mark the socket as non-blocking */

  IOCTL(socket_client, FIONBIO, &a) ;

  return socket_client ;
}

/*static int
print_data(owwl_conn *conn, owwl_data *data, void *user_data)
{
  char linebuf[128], namebuff[128] ;
  int length, unit_type = 0 ;
  FILE *fp ;
  
  fp = (FILE *) user_data ;
  
  if (data->str)
  {
    int arg = 0 ;
    
    while (arg >= 0)
    {
      fprintf(fp, "%-10s, %-10s, %10s, %s\n",
             owwl_name(data, namebuff, 128, &length, 0),
             owwl_arg_stem(data->device_type, data->device_subtype, arg),
             data->str(data, linebuf, 128, unit_type, -1, arg),
             owwl_unit_name(data, unit_type, arg)
      ) ;
      arg = owwl_next_arg(data, arg) ;
    }
    printf("\n") ;
  }
  
  return 0 ;
}*/

static void
tokenize_ignore(char *string)
{
  char *cp, *dp ;
  int found_end = 0 ;

  if (!string) return ;

  if (string[0] == '\0') return ;

  cp = dp = string ;

  while (!found_end)
  {
    char *copy, *copyi ;

    // Find end of this token
    while (strchr(",\0\n\t", *dp) == NULL) ++dp ;
    if (*dp == '\0') found_end = 1 ;
    *dp = '\0' ;
    ignored_vec = (char **) realloc(ignored_vec, sizeof(char **)*(ignore_count+1)) ;
    copyi = copy = strdup(cp) ;
    while (isspace(*copyi) && (*copyi != '\0')) ++copyi ;
    ignored_vec[ignore_count] = strdup(copyi) ;
    free(copy) ;
    ++ignore_count ;
    if (!found_end)
      cp = ++dp ;
  }
  return ;
}

static int
ignore_search(char *string)
{
  int i, j ;
  if (ignore_count == 0) return 0 ;
  for (i=0; i<ignore_count; ++i)
  {
    //if (0 == strcmp(string, ignored_vec[i])) return 1 ;
    // Special strcmp clone, that stops on * and skips #
    for (j=0; ((string[j]!='\0')&&(ignored_vec[i][j]!='\0')); ++j)
    {
      if (ignored_vec[i][j]=='*') return 1 ; // Wildcard
      if (ignored_vec[i][j]=='#') continue ; // Wildcard
      if (string[j]!=ignored_vec[i][j]) break ;
    }
    if ((string[j]=='\0')&&(ignored_vec[i][j]=='\0')) return 1 ; // Match
  }
  return 0 ;
}

/**
  * Print a data item to a string. Used for live file output.
  *
  *
  */
static int
print_str(owwl_conn *conn, owwl_data *data, void *user_data)
{
  char linebuf[128] ;
  int unit_type = 0 ;
  char *line ;

  //fp = (FILE *) user_data ;
  line = (char *) user_data ;

  if (data->str)
  {
    int arg = 0 ;

    while (arg >= 0)
    {
      data->str(data, linebuf, 128, unit_type, -1, arg) ;
      arg = owwl_next_arg(data, arg) ;
      if (line[0] != '\0') strcat(line, ", ") ;
      strcat(line, linebuf) ;
    }
  }

  return 0 ;
}

static void
check_conn(owwl_conn *conn)
{
  /* Kill off the client connexion */
  CLOSESOCK(conn->socket) ;
  //shutdown(conn->socket, SHUT_RDWR) ;

  /* Mark it as a dead socket */
  conn->socket = INVALID_SOCKET ;

  owwlog_warning("Link died\n") ;

#ifdef _WIN32
          Sleep(5000) ; /* Sleep for 5 s */
#else
          sleep(5) ; /* Sleep for 5 s */
#endif

}

/* Accumulate stats values for all channels */
static int
accumulate_stats(owwl_conn *conn, owwl_data *data, void *user_data)
{
  if (!data->user_data)
  {
	  // A new data struct is needed
	data_updated = 1; // This flag will cause the rrd template to be regenerated
    logger_build_data(conn, data, user_data) ;
#   ifdef HAVE_LIBGSL
    if (opt_caldir!=NULL)
      ((Logger *) data->user_data)->calib = calib_new(conn, data, opt_caldir) ;
#   endif
  }
  
# ifdef HAVE_LIBGSL
  if (((Logger *) data->user_data)->calib!=NULL)
    calib_eval(conn, data) ;
# endif
  
  logger_accumulate(conn, data, user_data) ;
  return 0 ; // Ok
}

struct name_id_assoc
{
  unsigned int id ;
  char *name ;
  int ignored ;
} ;

typedef struct name_id_assoc *NameIdPtr ;

/* Given a data stream and an arg, determine its id,
   and check this against the ignore list.
   Unlisted streams are added to the list, generating
   their name, and checking against the ignore list.
*/
static int
ignore_check(owwl_data *data, int arg)
{
  static NameIdPtr List = NULL ;
  static int ListSize = 0 ;
  char buffer[256] ;

  unsigned int Id ;
  int written ;
  int i ;

  if (!opt_ignore) return 0 ; // No ignore

  // What's the Id for this stream?
  Id = owwl_data_arg_to_id(data, arg) ;

  // Search List for Id
  for (i=0; i<ListSize; ++i)
  {
    if (Id == List[i].id)
    {
      // Found matching Id.
      // Is it on the ignore list?
      return List[i].ignored ;
    }
  }

  // Need fresh list entry
  ++ListSize ;
  List = realloc(List, ListSize * sizeof(struct name_id_assoc)) ;

  // Store id and name
  List[i].id = Id ;
  List[i].name = strdup(
                   owwl_arg_name(data, buffer, sizeof(buffer), &written, 0, arg)) ;

  // Check name against ignore list option
  List[i].ignored = ignore_search(List[i].name) ;

  return (List[i].ignored) ;
}

static int
print_header(owwl_conn *conn, owwl_data *data, void *user_data)
{
  int i, first = 0 ;
  Logger *log ;
  char buffer[32], *line ;
  int bufflen = 32 ;
  int written ;
  line = (char *) user_data ;

  log = (Logger *) data->user_data ;

  if (!log) return 0 ;

  if (line[0] == '\0') first = 1 ;

  for (i=0; i<log->results; ++i)
  {
    if (ignore_check(data, i)) continue ;

    //if (logger_result_name(data, user_data, i, buffer, bufflen) == LOGGER_OK)
    if (owwl_arg_name(data, buffer, bufflen, &written, 0, i) != NULL)
    {
      if (strlen(line) + written + ((first) ? 1 : 3) < MAXLINE)
      {
        if (!first) strcat(line, ", ") ;
        strcat(line, buffer) ;
      }
    }
    else
    {
      owwlog_warning("logger_result_name failed for arg %d\n", i) ;
      break ;
    }
  }

  return LOGGER_OK ; //logger_reset(conn, data, user_data) ;
}

static int
write_header(char *name, void *user_data)
{
  // File doesn't exist yet, so try to write headers

  char line[MAXLINE] ="" ;
  FILE *fp ;
  owwl_conn *conn ;

  conn = (owwl_conn *) user_data ;

  fp = fopen(name, "w") ;

  if (!fp) return FILE_CHECK_ERROR ;

  //lockf(fileno(fp), F_LOCK, 0);

  if (opt_iso_time)
    strncat(line, "Date/Time, ", MAXLINE) ;

  if (opt_unix_time)
    strncat(line, "Unix Time, ", MAXLINE-strlen(line)) ;

  owwl_foreach_ordered(conn, print_header, (void *) &line[strlen(line)]) ;
  strcat(line, "\n") ;

  if (fputs(line, fp) >= 0)
  {
    //lockf(fileno(fp), F_ULOCK, 0);
    fclose(fp) ;
    return FILE_CHECK_OK ;
  }
  //lockf(fileno(fp), F_ULOCK, 0);
  fclose(fp) ;
  return FILE_CHECK_ERROR ;
}

/* Print stats for a data stream */
static int
print_stats(owwl_conn *conn, owwl_data *data, void *user_data)
{
  int i, first = 0 ;
  Logger *log ;
  char buffer[32], *line ;
  int bufflen = 32 ;

  log = (Logger *) data->user_data ;
  line = (char *) user_data ;

  if (!log) return 0 ;

  if (line[0] == '\0') first = 1 ;

  if (logger_calculate(conn, data, user_data) == LOGGER_ERROR) return -1 ;

  for (i=0; i<log->results; ++i)
  {
    // Skip ignored streams
    if (ignore_check(data, i)) continue ;

    if (logger_result_string(data, user_data, i, buffer, bufflen, "%g") == LOGGER_OK)
    {
      if (strlen(line) + strlen(buffer) + ((first) ? 1 : 3) < MAXLINE)
      {
        if (!first) strcat(line, ", ") ;
        strcat(line, buffer) ;
      }
    }
    else
    {
      owwlog_warning("logger_result_string failed for arg %d\n", i) ;
      break ;
    }
  }

  return LOGGER_OK ; //logger_reset(conn, data, user_data) ;
}

/* Print stats for a data stream in RRDtool format */
static int
print_rrd_stats(owwl_conn *conn, owwl_data *data, void *user_data)
{
  int i, first = 0 ;
  Logger *log ;
  char buffer[32], *line ;
  int bufflen = 32 ;

  log = (Logger *) data->user_data ;
  line = (char *) user_data ;

  if (!log) return 0 ;

  if (line[0] == '\0') first = 1 ;

  if (logger_calculate(conn, data, user_data) == LOGGER_ERROR) return -1 ;

  for (i=0; i<log->results; ++i)
  {
    // Skip ignored streams
    if (ignore_check(data, i)) continue ;

    if (logger_result_string(data, user_data, i, buffer, bufflen, "%g") == LOGGER_OK)
    {
      if (strlen(line) + strlen(buffer) + ((first) ? 1 : 2) < MAXLINE)
      {
        if (!first) strcat(line, ":") ;
        strcat(line, buffer) ;
      }
    }
    else
    {
      owwlog_warning("logger_result_string failed for arg %d\n", i) ;
      break ;
    }
  }

  return LOGGER_OK ; //logger_reset(conn, data, user_data) ;
}

/* What is the directory to log to right now? */

static char *
dir_for_today(time_t t)
{
  static char todays_dir[MAXFILENAME] ;

  strncpy(todays_dir, opt_logdir, MAXFILENAME-1) ;
  todays_dir[MAXFILENAME-1] = '\0' ;

  strftime(&todays_dir[strlen(opt_logdir)],
           MAXFILENAME - strlen(opt_logdir) - 1,
           setup_format_logdir,
           gmtime(&t)) ;

  return todays_dir ;
}

/* What is the file to log to right now? */

static char *
file_for_today(time_t t)
{
  static char todays_file[MAXFILENAME] ;

  strncpy(todays_file, opt_logdir, MAXFILENAME-1) ;
  todays_file[MAXFILENAME-1] = '\0' ;

  strftime(&todays_file[strlen(opt_logdir)],
           MAXFILENAME - strlen(opt_logdir) - 1,
           setup_format_logfile,
           gmtime(&t)) ;

  return todays_file ;
}

/*
  process_new
 
  Run a new shell command (non-blocking)
*/
static int
process_new(char *command)
{
  pid_t newproc ;

  /* Ready to allocate new process */

#ifdef HAVE_FORK
  newproc = fork() ;

  if (newproc < 0)
  {
    /* Error from fork() */
    owwlog_warning("Failed to fork for command execution") ;
    return -1 ;
  }

  if (newproc <= 0)
  {
    /* We are in the child */
    /* Execute the command */
    char *argv[4] ;
    argv[0] = "sh";
    argv[1] = "-c";
    argv[2] = command ;
    argv[3] = 0;
    execve("/bin/sh", argv, NULL);

    /* We never reach here */
  }
#endif
  return 0 ;
}



static int
add_log_entry(owwl_conn *conn, void *user_data, char *log_name, char *log_entry, char mode[], time_t t)
{
  FILE *fp ;

  owwlog_message("add_log_entry %s\n%s\n",  log_name, log_entry) ;

  if (log_entry[0] == '\0')
  {
    owwlog_warning("Zero-length log line") ;
    return 1 ;
  }

  if (!log_name)
  {
    owwlog_warning("Tried to add entry to log file with NULL name") ;
    return -1 ;
  }

  if (log_name[0] == '\0')
  {
    owwlog_warning("Tried to add entry to log file with no name") ;
    return -1 ;
  }

  // If the file doesn't exist yet, try to write headers

  //if (file_check_file(log_name, 0) != FILE_CHECK_OK)
  //{
  //  char line[MAXLINE] ="" ;
  //
  //  if (!file_check_file(log_name, 0) != FILE_CHECK_OK)
  //    return -1 ; // Couldn't create new file
  //
  //  owwl_foreach(conn, print_header, (void *) line) ;
  //  strcat(line, "\n") ;
  //  add_log_entry(conn, user_data, log_name, line, mode, t) ;
  //}

  fp = fopen(log_name, mode) ;

  if (!fp)
  {
    owwlog_warning("Could not open file \"%s\"",
                   log_name) ;
    return -1 ;
  }

  //lockf(fileno(fp), F_LOCK, 0) ; // Get exclusive lock on log file
  if  (fputs(log_entry, fp) == EOF)
  {
    owwlog_warning("Error writing to \"%s\"", log_name) ;
    //lockf(fileno(fp), F_ULOCK, 0);
    fclose(fp) ;
    return -1 ;
  }

  //lockf(fileno(fp), F_ULOCK, 0);
  fclose(fp) ;

  // Execute user file if set
  if (opt_execute)
    process_new(opt_execute) ;

  return 0 ; // Ok
}

static void
write_log(owwl_conn *conn, void *user_data, char *name, int daily, time_t t)
{
  char logline[MAXLINE] ="" ;
  int offset ;

  owwlog_message("write_log %s\n", name) ;

  if (opt_iso_time)
    strftime(logline, MAXLINE, "%Y-%m-%d %H:%M:%S, ", gmtime(&t)) ;

  offset = strlen(logline) ;

  if (opt_unix_time)
    snprintf(&logline[offset], MAXLINE-offset, "%ld, ", t) ;

  offset = strlen(logline) ;

  owwl_foreach_ordered(conn, print_stats, (void *) &logline[offset]) ;
  strcat(logline, "\n") ;

  if (!daily)
  {
    if (file_check_file(name, 1, write_header, (void *) conn) != FILE_CHECK_OK)
    {
      owwlog_warning("File check failed on %s\n", name) ;
      return ; /* Failed */
    }

    if (add_log_entry(conn, user_data, name, logline, "a", t) != 0)
    {
      owwlog_warning("Attempt to update log (\"%s\") failed\n", name) ;
      return ; /* Failed */
    }
  }
  else
  {
    if (file_check_dir(name, 1) != FILE_CHECK_OK) return ; /* Failed */

    /* OK - main directory exists */
    if (file_check_dir(dir_for_today(t), 1) != FILE_CHECK_OK)
    {
      owwlog_warning("File check failed on directory %s\n", dir_for_today(t)) ;
      return ; /* Failed */
    }

    if (file_check_file(file_for_today(t), 1, write_header, (void *) conn) != FILE_CHECK_OK)
    {
      owwlog_warning("File check failed on file %s\n", file_for_today(t)) ;
      return ; /* Failed */
    }

    if (add_log_entry(conn, user_data, file_for_today(t), logline, "a", t) != 0)
    {
      owwlog_warning("Attempt to update log (\"%s\") failed [calling add_log_entry from log_start_log_event]", file_for_today(t)) ;
      return ; /* Failed */
    }
  }
}

/* Replace any non-alpha-numeric characters in name with '_' */
void
reduce_name(char *name)
{
  int j ;
  
  for (j=0; name[j]!='\0'; ++j)
  {
    if (!isalnum(name[j]))
      name[j] = '_' ;
  }
}

/* Fill string with colon-separated list of short stream names (no white space) */
static int
print_rrd_name(owwl_conn *conn, owwl_data *data, void *user_data)
{
  int i, first = 0 ;
  Logger *log ;
  char buffer[32], *line ;
  int bufflen = 32 ;
  int written ;
  line = (char *) user_data ;

  if (line[0] == '\0') first = 1 ;
  
  log = (Logger *) data->user_data ;

  if (!log) return 0 ;

  for (i=0; i<log->results; ++i)
  {
    if (ignore_check(data, i)) continue ;

    if (owwl_arg_name(data, buffer, bufflen, &written, 0, i) != NULL)
    {
      //int j ;
      
      if (strlen(line) + written + ((first) ? 1 : 2) < MAXLINE)
      {
        if (!first) strcat(line, ":") ;
        
        // Replace non-alnum characters with '_' in name
        reduce_name(buffer) ;
        strcat(line, buffer) ;
      }
    }
    else
    {
      owwlog_warning("logger_result_name failed for arg %d\n", i) ;
      break ;
    }
  }

  return LOGGER_OK ; //logger_reset(conn, data, user_data) ;
}

static int
build_rrd_names(char *name, void *user_data)
{
  // Build the rrd name template

  owwl_conn *conn ;

  conn = (owwl_conn *) user_data ;
  
  rrd_template[0] = '\0' ;

  owwl_foreach_ordered(conn, print_rrd_name, (void *) rrd_template) ;
  
  return 0 ;
}

static void
write_rrd(owwl_conn *conn, void *user_data, char *name, int daily, time_t t)
{
  char logline[MAXLINE] ="" ;
  int offset ;

  owwlog_message("write_rrd %s\n", name) ;

  offset = strlen(logline) ;

  snprintf(&logline[offset], MAXLINE-offset, "rrdtool update %s -t %s %ld:", name, rrd_template, t) ;

  offset = strlen(logline) ;

  owwl_foreach_ordered(conn, print_rrd_stats, (void *) &logline[offset]) ;
  strcat(logline, "\n") ;

  owwlog_message(logline) ;
  system(logline) ;
}

/* Set and check the log time */
static void
set_log_time(time_t suggested_time)
{
  log_time = suggested_time ;

  if ((opt_log_snap) && (log_time % opt_time != 0))
    log_time -= log_time % opt_time ;

  if (log_time + opt_time <= the_time)
  {
    /* Need to catch up */
    log_time += opt_time *
                ((the_time - log_time) / opt_time) ;
  }
}

#ifndef HAVE_DAEMON

#ifdef HAVE_UNISTD_H
//#include <unistd.h>
//#include <stdlib.h>
#include <fcntl.h>
//#include <signal.h>
#include <sys/types.h>
//#include <sys/wait.h>
//#include <errno.h>

/* closeall() -- close all FDs >= a specified value */

#ifdef HAVE_FORK
static void closeall(int fd)
{
  /*int fdlimit = sysconf(_SC_OPEN_MAX);*/

  while (fd < 64)
    close(fd++);
}

/* daemon() - detach process from user and disappear into the background
 * returns -1 on failure, but you can't do much except exit in that case
 * since we may already have forked. This is based on the BSD version,
 * so the caller is responsible for things like the umask, etc.
 */

/* believed to work on all Posix systems */

int daemon(int nochdir, int noclose)
{
  switch (fork())
  {
  case 0:  break;
  case -1: return -1;
  default: _exit(0);          /* exit the original process */
  }

  if (setsid() < 0)               /* shoudn't fail */
    return -1;

  /* dyke out this switch if you want to acquire a control tty in */
  /* the future -- not normally advisable for daemons */

  switch (fork())
  {
  case 0:  break;
  case -1: return -1;
  default: _exit(0);
  }

  if (!nochdir)
    chdir("/");

  if (!noclose)
  {
    closeall(0);
    open("/dev/null",O_RDWR);
    dup(0); dup(0);
  }

  return 0;
}
#else /* ifdef HAVE_FORK */
int daemon(int nochdir, int noclose)
{ return -1 ; }
#endif /* ifdef HAVE_FORK */
#else
int daemon(int nochdir, int noclose)
{ return -1 ; }

#endif /* ifdef HAVE_UNISTD_H */

#endif /* ifndef HAVE_DAEMON */

/*static int
save_pid(void)
{
  FILE *fp ;
  
  fp = fopen(".owwlog.pid", "w") ;
  if (!fp) return -1 ;
  fprintf(fp, "%ld", (long) getpid()) ;
  fclose(fp) ;
  return 0 ;
}
 
static int
load_pid(pid_t *pid)
{
  FILE *fp ;
  
  fp = fopen(".owwlog.pid", "r") ;
  if (!fp) return -1 ;
  fscanf(fp, "%ld", (long *) pid) ;
  fclose(fp) ;
  return 0 ;
}*/

static void
create_server(owwl_conn *conn, int port)
{
  struct sockaddr *addr = NULL ;
  struct sockaddr_in addr_in ;
  int s, i=1 ;
  unsigned long a = 1L;
  int addrLength;

  addr = (struct sockaddr *) &addr_in ;
    
  s = socket (PF_INET, SOCK_STREAM, 0) ;
    
  setsockopt (s, SOL_SOCKET, SO_REUSEADDR, (char *) &i, sizeof (i));

  IOCTL (s, FIONBIO, &a);

  addr_in.sin_port = htons (port);
  addr_in.sin_family = AF_INET ;
  memset (&addr_in.sin_addr, 0, sizeof (addr_in.sin_addr));
  addrLength = sizeof (addr_in);

  if (bind (s,
      addr,
      sizeof (addr_in)))
  {
    owwlog_warning("Unable to claim TCP port %d. Perhaps another task claimed it already?\n", port);
    CLOSESOCK(s);
    return;
  }

  if (listen (s, 5))
  {
    owwlog_warning("listen() failed, TCP port %d\n", port);
    CLOSESOCK(s);
    return;
  }
  
  owwlog_message("Listening socket created on port %d\n", port) ;
  
  owwl_tx_new_server(conn, s) ;
}

int main(int argc, char *argv[])

{
  int s, r ;
  //int give_up = 0 ;
  time_t last = 0 ;

  owwl_conn *conn ;
  //owwl_buffer buff = Owwl_Buffer_Init ;

  time_t next_log ;
  
  poptContext optCon;   /* context for parsing command-line options */
  struct poptOption optionsTable[] =
    {
      { "abort", 'a', POPT_ARG_INT, (void *) &opt_abort, 0,
        "Abort timeout / retry interval", "time in seconds" },
      { "Broadcaster", 'B', POPT_ARG_NONE, (void *) &opt_broadcaster, 0,
        "Run as a re-broadcasting server", NULL },
      { "broadcast", 'b', POPT_ARG_INT, (void *) &opt_broadcast, 0,
        "Port number to run a re-broadcasting server", "port number" },
#     ifdef HAVE_LIBGSL
      { "Calibration", 'C', POPT_ARG_STRING, (void *) &opt_caldir, 0,
        "Use files in this directory for calibration", "directory path"
      },
#     endif
      { "Directory", 'D', POPT_ARG_STRING, (void *) &opt_logdir, 0,
        "Log to files in this directory", "directory path"
      },
      { "daemon", 'd', POPT_ARG_NONE, (void *) &opt_daemon, 0,
        "Run in daemon mode", NULL },
      { "execute", 'e', POPT_ARG_STRING, (void *) &opt_execute, 0,
        "Execute file after log updates", "file" },
      { "ISO", 'I', POPT_ARG_NONE, (void *) &opt_iso_time, 0,
        "Print ISO 8601 time values", NULL },
      { "Ignore", 'i', POPT_ARG_STRING, (void *) &opt_ignore, 0,
        "Ignore data streams", "stream list" },
      { "Live", 'L', POPT_ARG_STRING, (void *) &opt_livefile, 0,
        "Log live updates to this file", "file" },
      { "Log", 'l', POPT_ARG_STRING, (void *) &opt_logfile, 0,
        "Log to this file", "log path" },
      { "port", 'p', POPT_ARG_INT, (void *) &opt_port, 0,
        "Port number to connect on", "port number" },
      { "quiet", 'q', POPT_ARG_NONE, (void *) &opt_quiet, 0,
        "Don't print messages after startup", NULL },
      { "rrdfile", 'r', POPT_ARG_STRING, (void *) &opt_rrdfile, 0,
        "Update this RRD database", "file" },
      { "Snap", 'S', POPT_ARG_NONE, (void *) &opt_log_snap, 0,
        "Snap log times", NULL },
      { "server", 's', POPT_ARG_STRING, (void *) &opt_server, 0,
        "Server to contact", "server" },
      { "time", 't', POPT_ARG_INT, (void *) &opt_time, 0,
        "Log interval", "time in seconds" },
      { "unix", 'u', POPT_ARG_NONE, (void *) &opt_unix_time, 0,
        "Print Unix time values", NULL },
      { "version", 'v', POPT_ARG_NONE, NULL, 'v',
        "Print version information", NULL },
        POPT_AUTOHELP
        POPT_TABLEEND
    };
  int arg ;
  int done=0;

  time(&the_time) ;

  optCon = poptGetContext(
             NULL,
             argc,
             (const char **) argv,
             (const struct poptOption *) optionsTable,
             0);
  poptSetOtherOptionHelp(optCon, "[OPTIONS]* <port>");

  while ((arg = poptGetNextOpt(optCon)) >= 0)
  {
    switch (arg)
    {
      case 'v': /* Print version info */
#ifdef VERSION
        printf("owwlog version %s\n", VERSION) ;
        printf("Simon Melhuish\n") ;
#endif
        exit(0) ;
        break ;
    }
  }

  if (arg < -1)
  {
    /* an error occurred during option processing */
    owwlog_error("%s: %s",
                 poptBadOption(optCon, POPT_BADOPTION_NOALIAS),
                 poptStrerror(arg));
    poptFreeContext(optCon) ;
    return 1;
  }
  
  poptFreeContext(optCon) ;

  if (opt_daemon)
    owwlog_message("owwlog will run in daemon mode\n") ;

  if (opt_logfile && opt_logdir)
  {
    owwlog_error("You cannot log both to a file and a directory\n") ;
    //exit(0) ;
  }

  if (opt_logfile)
    owwlog_message("Logging to single file \"%s\" every %d seconds\n", opt_logfile, opt_time) ;

  if (opt_logdir)
    owwlog_message("Logging to daily files in directory \"%s\" every %d seconds\n", opt_logdir, opt_time) ;

  if (opt_ignore)
  {
    int i ;
    tokenize_ignore(opt_ignore) ;
    owwlog_message("Ignoring channels:\n") ;
    for (i=0; i<ignore_count; ++i)
      owwlog_message("\t%s\n", ignored_vec[i]) ;
  }
  
  if (opt_caldir)
  {
    owwlog_message("Files in %s will be used for calibration\n", opt_caldir) ;
  }

  next_log = time(NULL) + opt_time ;

#ifndef _WIN32
  setlinebuf(stdout) ;
#endif

  if (!opt_server)
  {
    owwlog_error("You must name a server\n") ;
    //exit(0) ;
  }

#ifdef _WIN32
  WSADATA wsadata;
  int rc ;

  rc = WSAStartup(2, &wsadata);
  if(rc)
  {
    owwlog_error("WSAStartup FAILED: err=%d\n", GetLastError());
    //return 1;
  }
#endif

  s = client_connect(opt_server, opt_port) ;
  last = time(NULL) ;

  if (s <= 0)
  {
    owwlog_warning("Unable to establish connexion\n") ;
  }

  conn = owwl_new_conn(s, NULL) ;

  if (conn == NULL)
  {
    owwlog_error("Null conn\n") ;
  }

  owwlog_message("conn created\n") ;

  if (opt_daemon)
  {
    owwlog_message("Changing to daemon mode\n") ;
    if (-1 == daemon (1 /* Don't chdir */, 0 /* Do close stdio */))
    {
      owwlog_error("daemon failed\n") ;
      //exit(1) ;
    }
  }
  
  // Should we start our own server?
  
  if (opt_broadcaster)
    create_server(conn, opt_broadcast) ;

  //save_pid() ;

  time(&log_time) ;
  
  if (opt_quiet) run_quiet = 1 ;

  while (1)
  {

#ifdef HAVE_WAIT3
    wait3(NULL, WNOHANG, NULL) ; // Wait on any children
#endif
    time(&the_time) ;

    if (conn->socket > 0)
    {
      fd_set fds ;
      fd_set fde ;
      int poll_result ;
      struct timeval tv = {0, 100} ;
      FD_ZERO(&fds) ;
      FD_SET(conn->socket, &fds) ;
      FD_ZERO(&fde) ;
      FD_SET(conn->socket, &fde) ;
      poll_result = select(conn->socket+1, &fds, NULL, &fde, &tv) ;
      if ((poll_result > 0) && FD_ISSET(conn->socket, &fde))
      {
        /* Exception on socket - Help! */
        owwlog_warning("Socket error: %d - %s\n", LASTSOCKERR, strerror(LASTSOCKERR));
        check_conn(conn) ;
        last = time(NULL) ;
      }
      else if ((poll_result > 0) && FD_ISSET(conn->socket, &fds))
      {

        r=owwl_read(conn) ;
        //printf("read -> %d\n", r) ;
        switch(r)
        {
        case Owwl_Read_Read_And_Decoded:
          {
            owwl_tx_poll_servers(conn) ;
            data_updated = 0; // Assume no data structs were added to conn - accumulate_stats will pick up anything new
            owwl_foreach(conn, accumulate_stats, (void *) opt_caldir) ;
            //owwl_tx_build(conn, OWW_TRX_MSG_WSDATA, &buff) ;
            //owwl_tx(conn, &buff) ;
            last = time(NULL) ;

            if (opt_livefile)
            {
              FILE *fp ;
              fp = fopen(opt_livefile, "w") ;
              if (!fp)
              {
                owwlog_warning("Unable to open live update file \"%s\"", opt_livefile) ;
              }
              else
              {
                char logline[MAXLINE] ="" ;

                if (opt_iso_time)
                {
                  strftime(logline, MAXLINE, "%Y-%m-%d %H:%M:%S", gmtime(&the_time)) ;
                  //fputs(logline, fp) ;
                }

                if (opt_unix_time)
                  sprintf(logline, "%ld", the_time) ;

                //owwl_foreach_ordered(conn, print_data, (void *) fp) ;
                //printf("%ld\n\n", conn->data_time) ;
                owwl_foreach_ordered(conn, print_str, (void *) logline) ;
                strcat(logline, "\n") ;
                //lockf(fileno(fp), F_LOCK, 0);
                fputs(logline, fp) ;
                //lockf(fileno(fp), F_ULOCK, 0);
                fclose(fp) ;
              }
            }


            if ((opt_logdir || opt_logfile || opt_rrdfile) && (the_time >= log_time + opt_time)) // Time for log update?
            {
              set_log_time(log_time + opt_time) ;
              /*next_log += opt_time ;
              if (last >= next_log)
                next_log = last + opt_time ;*/
                
              if (opt_rrdfile)
              {
                if ((rrd_template[0]=='\0') || data_updated)
                  build_rrd_names(NULL, conn) ;
                
                write_rrd(conn, NULL, opt_rrdfile, 0, the_time) ;
              }

              if (opt_logdir)
                write_log(conn, NULL, opt_logdir, 1, the_time) ;
              else if (opt_logfile)
                write_log(conn, NULL, opt_logfile, 0, the_time) ;
              owwl_foreach_ordered(conn, logger_reset, NULL) ;
            }
            
            if (opt_broadcaster)
            {
              owwl_buffer buff = Owwl_Buffer_Init ;
            
              owwl_tx_build(conn, OWW_TRX_MSG_WSDATA, &buff) ;
              owwl_tx(conn, &buff) ;
              free((void *) buff.buffer) ;
            }

            break ;
          }

        case Owwl_Read_Again:
          owwl_tx_poll_servers(conn) ;
#ifdef _WIN32
          Sleep(10) ; /* Sleep for 10 ms */
#else
          usleep(10000) ; /* Sleep for 10 ms */
#endif
          /* Check for timeout - 2 intervals + opt_abort (default 30 s) */
          if ((last != 0) && (time(NULL) > last + conn->interval * 2 + opt_abort))
          {
            owwlog_warning("Timeout\n") ;
            check_conn(conn) ;
            last = time(NULL) ;
          }
          break ;

        case Owwl_Read_Disconnect:
          /* Try to reconnect */
          owwlog_warning("Server died\nAuto-reconnect\n") ;
          CLOSESOCK(conn->socket) ;
          conn->socket = client_connect(opt_server, opt_port) ;
          if (conn->socket > 0)
          {
            last = time(NULL) ;
            owwlog_message("Reconnected\n") ;
          }
          else
          {
            owwlog_warning("Reconnect failed\n") ;
          }
          break ;

        case Owwl_Read_Error:
          owwlog_warning("Read error: %d - %s\n", LASTSOCKERR, strerror(LASTSOCKERR));
          check_conn(conn) ;
          last = time(NULL) ;
          break ;
        }
      }
      else
      {
        /* No data ready */
        /* Check for client connexions */
        owwl_tx_poll_servers(conn) ;
        
        /* Check for timeout - 2 intervals + opt_abort (default 30 s) */
        if ((last != 0) && (time(NULL) > last + conn->interval * 2 + opt_abort))
        {
          owwlog_warning("Timeout\n") ;
          check_conn(conn) ;
          last = time(NULL) ;
        }
      }
    }
    else /* Not connected for some reason */
    {
      /* Try to reconnect */
      owwlog_warning("Reconnecting\n") ;
      conn->socket = client_connect(opt_server, opt_port) ;
      if (conn->socket <= 0) /* Failed to connect? */
      {
        owwlog_warning("Reconnexion failed\n") ;
#ifdef _WIN32
        Sleep(1000 * opt_abort) ;
#else
        sleep(opt_abort) ;
#endif
        last = 0 ;
      }
      else
      {
        owwlog_message("Reconnected\n") ;
        last = time(NULL) ;
      }
    }
  }

  owwl_free(conn) ;

  owwlog_message("conn freed\n") ;

#ifdef _WIN32
  WSACleanup();
#endif

  return (0);
}
